"""Test static methods."""

from pathlib import Path

from treestamps.tree import Treestamps

__all__ = ()


class TestStaticMethods:
    """Test static methods."""

    def test_get_dir(self):
        """Test dirpath."""
        file_path = Path(__file__)
        dir_path = file_path.parent
        assert Treestamps.get_dir(dir_path) == dir_path
        assert Treestamps.get_dir(file_path) == dir_path

    def test_maxnone(self):
        """Test maxnone."""
        assert Treestamps.max_none(1, 2) == 2  # noqa: PLR2004
        assert Treestamps.max_none(None, 2) == 2  # noqa: PLR2004
        assert Treestamps.max_none(1, None) == 1
        assert Treestamps.max_none(None, None) is None



"""Print Messages."""

from termcolor import cprint


class Printer:
    """Printing messages."""

    def __init__(self, verbose: int):
        """Initialize verbosity and flags."""
        self._verbose = verbose
        self._after_newline = True

    def _message(
        self, reason, color="white", attrs=None, *, force_verbose=False, end="\n"
    ):
        """Print a dot or skip message."""
        if self._verbose < 1:
            return
        if (self._verbose == 1 and not force_verbose) or not reason:
            cprint(".", color, attrs=attrs, end="", flush=True)
            self._after_newline = False
            return
        if not self._after_newline:
            reason = "\n" + reason
        attrs = attrs if attrs else []
        cprint(reason, color, attrs=attrs, end=end, flush=True)
        if end:
            self._after_newline = True

    def skip(self, message, path):
        """Skip Message."""
        parts = ["Skip", message, str(path)]
        message = ": ".join(parts)
        self._message(message, color="dark_grey")

    def load(self, message, path):
        """Save timestamps."""
        message = f"{message} {path}"
        self._message(message, color="cyan")

    def save(self, message, path):
        """Save timestamps."""
        message = f"{message} {path}"
        self._message(message, color="green", attrs=["bold"])

    def compact(self, message, path, timestamp):
        """Compact timestamps."""
        message = ": ".join((message, str(path), str(timestamp)))
        self._message(message, color="dark_grey", attrs=["dark"])

    def warn(self, message: str, exc: Exception | None = None):
        """Warning."""
        message = "WARNING: " + message
        if exc:
            message += f": {exc}"
        self._after_newline = False
        self._message(message, color="light_yellow", force_verbose=True)

    def error(self, message: str, exc: Exception | None = None):
        """Error."""
        message = "ERROR: " + message
        if exc:
            message += f": {exc}"
        self._message(message, color="light_red", force_verbose=True)



"""Treestamps Config methods."""

from abc import ABC
from collections.abc import Iterable, Mapping
from dataclasses import dataclass

from ruamel.yaml.comments import CommentedMap, CommentedSet


@dataclass
class CommonConfig(ABC):
    """Common Config meant to be subclassed."""

    program_name: str
    verbose: int = 0
    symlinks: bool = True
    ignore: Iterable[str] = frozenset()
    check_config: bool = True
    program_config: Mapping | None = None
    program_config_keys: Iterable[str] = frozenset()

    @classmethod
    def normalize_config(cls, value):
        """Recursively convert iterables into frozen sorted unique lists."""
        if isinstance(value, Mapping | CommentedMap):
            value = dict(
                sorted(
                    (key, cls.normalize_config(sub_value))
                    for key, sub_value in value.items()
                )
            )
        if isinstance(value, list | tuple):
            value = tuple(sorted(cls.normalize_config(e) for e in value))
        if isinstance(value, set | CommentedSet):
            value = frozenset(cls.normalize_config(e) for e in value)
        return value

    def __post_init__(self):
        """Fix types and normalize program config dict."""
        self.ignore = frozenset(self.ignore)
        self.program_config_keys = frozenset(self.program_config_keys)

        # Filter dict by keys
        if self.program_config:

            def filter_func(pair):
                return pair[0] in self.program_config_keys

            program_config = (
                dict(filter(filter_func, self.program_config.items()))
                if self.program_config_keys
                else self.program_config
            )
            self.program_config = program_config

        self.program_config = self.normalize_config(self.program_config)



"""A dict of Treestamps."""

from collections.abc import Iterable, Mapping
from copy import copy
from dataclasses import asdict, dataclass
from pathlib import Path
from warnings import warn

from treestamps.config import CommonConfig
from treestamps.printer import Printer
from treestamps.tree import Treestamps
from treestamps.tree.init import TreestampsConfig


@dataclass
class GrovestampsConfig(CommonConfig):
    """Grovestamps config."""

    paths: Iterable[str | Path] = ()

    def __post_init__(self):
        """
        Pathify, filter, dedupe, order and tuplify paths.

        This order creates dir based treestamps before files so dirs get children
        recursed and files only don't.
        """
        super().__post_init__()
        dirs: set[Path] = set()
        files: set[Path] = set()
        for path_str in self.paths:
            path = Path(path_str)
            if not self.symlinks and path.is_symlink():
                continue
            if path.is_dir():
                dirs.add(path)
            else:
                files.add(path)
        self.paths = tuple(sorted(dirs) + sorted(files))

    def get_treestamps_config_dict(self):
        """Get a treestamps style config dict from this config."""
        config = copy(self)
        if config.program_config is not None:
            config.program_config = dict(config.program_config)
        config_dict = asdict(config)
        config_dict.pop("paths", None)
        return config_dict


class Grovestamps(dict):
    """A path keyed dict of Treestamps."""

    def __init__(self, config: GrovestampsConfig) -> None:
        """Create a dictionary of Treestamps keyed with paths."""
        self._config = config
        self._printer = Printer(config.verbose)

        treestamps_config_dict = self._config.get_treestamps_config_dict()

        for top_path in self._config.paths:
            root_dir = Treestamps.get_dir(top_path)
            if root_dir in self:
                continue
            tree_config = TreestampsConfig(
                **treestamps_config_dict, path=Path(top_path)
            )
            ts = Treestamps(tree_config, self._printer)
            ts.loadf_tree()
            self[root_dir] = ts

        self.filename = Treestamps.get_filename(self._config.program_name)
        self.wal_filename = Treestamps.get_wal_filename(self._config.program_name)

    def load(self, path: str | Path, yaml: Mapping | str | bytes | Path):
        """Load a timestamp yaml dict into the correct treestamps."""
        path = Path(path)
        if not path.is_dir():
            path = path.parent
        for top_path in self.keys():
            if path.is_relative_to(top_path):
                treestamps = self[top_path]
                if isinstance(yaml, Mapping):
                    treestamps.load_dict(yaml)
                elif isinstance(yaml, str | bytes):
                    treestamps.loads(path, yaml)
                elif isinstance(yaml, Path):
                    treestamps.loadf(path)
                break
        else:
            reason = f"load dict to {path} is not relative to any Grovetamps path: {tuple(self.keys())}"
            raise ValueError(reason)

    def load_map(self, grove: Mapping[Path, Mapping | str | bytes | Path]):
        """Load a grove of treestamps from a mapping."""
        for path, yaml in grove.items():
            self.load(path, yaml)

    def loads(self, path: str | Path, yaml_str: str):
        """Load a timestamp yaml string into the correct treestamps."""
        self.load(path, yaml_str)

    def loadf(self, path: str | Path) -> None:
        """Load a timestamp file into the correct treestamps."""
        path = Path(path)
        self.load(path.parent, path)

    def dumpf(self) -> None:
        """Dump all treestamps."""
        for top_path, treestamps in self.items():
            self._printer.save("Saving timestamps for", top_path)
            treestamps.dumpf()

    def dumps(self) -> dict[Path, str]:
        """Dump all treestamps to dict as strings."""
        return {top_path: treestamps.dumps() for top_path, treestamps in self.items()}

    def dump_dict(self) -> dict[Path, dict]:
        """Dump all treestamps to dict."""
        return {
            top_path: treestamps.dump_dict() for top_path, treestamps in self.items()
        }

    def dump(self) -> None:
        """Alias for dumpf."""
        warn("replaced by Grovestamps.dumpf()", PendingDeprecationWarning, stacklevel=2)
        self.dumpf()
